#!/usr/bin/env node
"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const fs_1 = require("fs");
const path_1 = __importDefault(require("path"));
const ProgressBar_1 = require("./ProgressBar");
const AWSS3Glacier_1 = require("./service/AWSS3Glacier");
const BackblazeB2_1 = require("./service/BackblazeB2");
const upload_1 = require("./upload/upload");
const nullableReadFile_1 = require("./util/nullableReadFile");
const minimist = require("minimist");
const DEFAULT_CONCURRENT_UPLOADS = 3;
const SERVICES = {
    aws: {
        service: AWSS3Glacier_1.AWSS3Glacier,
        options: AWSS3Glacier_1.parseAWSS3GlacierOptions,
    },
    b2: {
        service: BackblazeB2_1.BackblazeB2,
        options: BackblazeB2_1.parseBackblazeB2Options,
    },
};
const exit = (error) => {
    if (error) {
        console.error(error);
    }
    process.exit(error ? 1 : 0);
    // This is for TypeScript, which doesn't recognise that
    // calling process.exit doesn't return.
    throw 1;
};
const sessionPath = (workDir) => {
    return path_1.default.join(workDir, "session");
};
const statePath = (workDir, key) => {
    return path_1.default.join(workDir, `state_${key}`);
};
const args = minimist(process.argv.slice(2));
const concurrentUploads = +args.concurrency || DEFAULT_CONCURRENT_UPLOADS;
const quiet = !!args.quiet;
const verbose = !!args.verbose;
const force = !!args.force;
const { service, options: parseOptions } = SERVICES[args.service.toLowerCase()];
const serviceOptions = parseOptions(args);
const progressBar = quiet ? null : new ProgressBar_1.ProgressBar(":title [:bar] :percent%");
const logFormat = (msg) => `[${new Date().toISOString()}] ${msg}`;
const filePath = fs_1.realpathSync(args.file);
const workingDirectory = fs_1.realpathSync(args.work);
let fileStats;
const ctx = {
    concurrentUploads,
    file: {
        path: filePath,
        // File stats aren't available right now, but will be loaded later before $ctx is used.
        get size() {
            return fileStats.size;
        },
        get lastModified() {
            return fileStats.mtime.toISOString();
        },
    },
    force,
    log: (msg, info = false) => {
        if (info && !verbose) {
            return;
        }
        const out = logFormat(msg);
        if (progressBar) {
            progressBar.log(out);
        }
        else {
            console.error(out);
        }
    },
    readState: key => nullableReadFile_1.nullableReadFile(statePath(workingDirectory, key)),
    resumeSession: () => nullableReadFile_1.nullableReadJson(sessionPath(workingDirectory)),
    updateProgress: p => progressBar && progressBar.update({
        percent: p.completeRatio * 100,
        title: p.description,
    }),
    writeSession: session => fs_1.promises.writeFile(sessionPath(workingDirectory), JSON.stringify(session)),
    writeState: (key, value) => fs_1.promises.writeFile(statePath(workingDirectory, key), value),
};
const onEnd = (err) => {
    if (progressBar) {
        progressBar.clear();
    }
    if (err) {
        exit(err.stack || err);
    }
    else {
        exit();
    }
};
Promise.all([filePath, workingDirectory].map(fs_1.promises.stat))
    .then(([file, work]) => {
    if (!file.isFile()) {
        throw new TypeError(`${filePath} is not a file`);
    }
    if (!work.isDirectory()) {
        throw new TypeError(`${workingDirectory} is not a directory`);
    }
    fileStats = file;
    return service.fromOptions(serviceOptions);
})
    .then(serviceState => upload_1.upload(ctx, service, serviceState))
    .then(onEnd, onEnd);
//# sourceMappingURL=main.js.map