"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const aws_sdk_1 = __importDefault(require("aws-sdk"));
const crypto_1 = __importDefault(require("crypto"));
const fs_1 = require("fs");
const assert_1 = require("../util/assert");
const nextPowerOfTwo_1 = require("../util/nextPowerOfTwo");
const isReadStreamWithPartDetails = (body) => {
    return body instanceof fs_1.ReadStream && !!body.__ltsuPartDetails;
};
const joinReadStreamWithPartDetails = (stream, part) => {
    return Object.assign(stream, { __ltsuPartDetails: part });
};
const optionalMap = (val, mapper) => {
    return val == undefined ? undefined : mapper(val);
};
exports.parseAWSS3GlacierOptions = (args) => {
    const region = optionalMap(args.region, v => v.toString());
    const accessId = optionalMap(args.access, v => v.toString());
    const secretKey = optionalMap(args.secret, v => v.toString());
    const vaultName = args.vault.toString();
    return {
        region, accessId, secretKey, vaultName,
    };
};
const buildFinalTreeHash = (hashes) => {
    while (hashes.length > 1) {
        const newHashes = [];
        for (let i = 0; i < hashes.length - 1; i += 2) {
            newHashes.push(crypto_1.default.createHash("sha256")
                .update(Buffer.concat(hashes.slice(i, i + 2)))
                .digest());
        }
        if (hashes.length % 2 === 1) {
            newHashes.push(hashes[hashes.length - 1]);
        }
        hashes = newHashes;
    }
    return hashes[0];
};
const MiB = 1024 * 1024;
// This is difficult, as it needs to build the tree hash while reading the part,
// not collect hashes for the entire part first (which might result in heavy
// memory usage for large part sizes).
const calculateTreeAndLinearHashesOfPart = ({ path, start, end }) => __awaiter(this, void 0, void 0, function* () {
    const tree = [];
    const linear = crypto_1.default.createHash("sha256");
    const fd = yield fs_1.promises.open(path, "r");
    try {
        for (let chunkStart = start; chunkStart <= end; chunkStart += MiB) {
            // Calculate chunk range and length.
            const chunkEnd = Math.min(chunkStart + MiB - 1, end);
            const chunkSize = chunkEnd - chunkStart + 1;
            // Read chunk.
            const chunk = Buffer.allocUnsafe(chunkSize);
            const { bytesRead } = yield fd.read(chunk, 0, chunkSize, chunkStart);
            assert_1.assertTrue(bytesRead === chunkSize);
            // Build linear hash.
            linear.update(chunk);
            // Build tree hash.
            tree.push({
                level: 1,
                hash: crypto_1.default.createHash("sha256").update(chunk).digest(),
            });
            while (tree.length > 1 && tree[tree.length - 1].level === tree[tree.length - 2].level) {
                const right = assert_1.assertExists(tree.pop());
                const left = assert_1.assertExists(tree.pop());
                tree.push({
                    level: right.level + 1,
                    hash: crypto_1.default.createHash("sha256")
                        .update(Buffer.concat([left.hash, right.hash]))
                        .digest(),
                });
            }
        }
    }
    finally {
        yield fd.close();
    }
    // Finalise tree hash.
    while (tree.length > 1) {
        const right = assert_1.assertExists(tree.pop());
        const left = assert_1.assertExists(tree.pop());
        tree.push({
            // To silence type errors.
            level: -1,
            hash: crypto_1.default.createHash("sha256")
                .update(Buffer.concat([left.hash, right.hash]))
                .digest(),
        });
    }
    return {
        tree: tree[0].hash,
        linear: linear.digest(),
    };
});
// Patch internal method addTreeHashHeaders on AWS.Glacier class to support
// a body of type `fs.ReadStream & {psf: PartStreamFactory}`. By default, the
// SDK tries to convert the ReadStream into a Buffer, which fails as Buffer.from
// doesn't support an argument of type fs.ReadStream. In addition, we don't want
// to read the entire part (which could be up to 4 GiB) into memory, so we pass
// a ReadStream as the body value (which the SDK does support uploading from) with
// a `psf` property that, when called, creates a new ReadStream of the same part,
// as the initial ReadStream will be consumed when building the hashes.
// @ts-ignore
aws_sdk_1.default.Glacier.prototype.addTreeHashHeaders = (request, callNextListener) => __awaiter(this, void 0, void 0, function* () {
    const body = request.params.body;
    if (isReadStreamWithPartDetails(body)) {
        const { path, start, end } = body.__ltsuPartDetails;
        const { linear, tree } = yield calculateTreeAndLinearHashesOfPart({ path, start, end });
        // The property "X-Amz-Content-Sha256" needs to be exactly in that letter casing,
        // otherwise the SDK will think it doesn't exist and try to hash the ReadStream (which fails).
        request.httpRequest.headers["X-Amz-Content-Sha256"] = linear.toString("hex");
        // TODO Replace with assertion
        if (request.httpRequest.headers["x-amz-sha256-tree-hash"]) {
            throw new Error(`Tree hash header already exists on AWS Glacier request`);
        }
        request.httpRequest.headers["x-amz-sha256-tree-hash"] = tree.toString("hex");
        // Length property is needed as by default the SDK gets the length of a fs.ReadStream by the size of the
        // file at `stream._path`, which is the length of the entire file.
        request.params.body = Object.assign(body, { length: end - start + 1 });
    }
    else if (body != undefined) {
        // Default code.
        const { linearHash, treeHash } = request.service.computeChecksums(body);
        request.httpRequest.headers["X-Amz-Content-Sha256"] = linearHash;
        if (!request.httpRequest.headers["x-amz-sha256-tree-hash"]) {
            request.httpRequest.headers["x-amz-sha256-tree-hash"] = treeHash;
        }
    }
    callNextListener();
});
// The AWS SDK uses this to determine that the function is asynchronous and so will provide
// a callback and wait for it to be called with any error before continuing.
// This is necessary as our calculateTreeAndLinearHashesOfPart function is asynchronous.
// @ts-ignore
aws_sdk_1.default.Glacier.prototype.addTreeHashHeaders._isAsync = true;
exports.AWSS3Glacier = {
    // See https://docs.aws.amazon.com/amazonglacier/latest/dev/uploading-archive-mpu.html for limits.
    minParts: 1,
    maxParts: 10000,
    minPartSize: 1024 * 1024,
    maxPartSize: 1024 * 1024 * 1024 * 4,
    fromOptions(options) {
        return __awaiter(this, void 0, void 0, function* () {
            return {
                service: new aws_sdk_1.default.Glacier({
                    accessKeyId: options.accessId || undefined,
                    secretAccessKey: options.secretKey || undefined,
                    region: options.region || undefined,
                }),
                vaultName: options.vaultName,
            };
        });
    },
    completeUpload(s, uploadId, fileSize, partHashes) {
        return __awaiter(this, void 0, void 0, function* () {
            yield s.service.completeMultipartUpload({
                accountId: "-",
                archiveSize: `${fileSize}`,
                checksum: buildFinalTreeHash(partHashes).toString("hex"),
                uploadId: uploadId,
                vaultName: s.vaultName,
            }).promise();
        });
    },
    idealPartSize(_, fileSize) {
        return __awaiter(this, void 0, void 0, function* () {
            return nextPowerOfTwo_1.nextPowerOfTwo(fileSize / this.maxParts);
        });
    },
    initiateNewUpload(s, fileName, partSize) {
        return __awaiter(this, void 0, void 0, function* () {
            const result = yield s.service.initiateMultipartUpload({
                accountId: "-",
                vaultName: s.vaultName,
                partSize: `${partSize}`,
                archiveDescription: fileName,
            }).promise();
            return assert_1.assertExists(result.uploadId);
        });
    },
    uploadPart(s, uploadId, part) {
        return __awaiter(this, void 0, void 0, function* () {
            const { path, start, end } = part;
            const res = yield s.service.uploadMultipartPart({
                accountId: "-",
                // See comment above the AWS.Glacier.prototype.addTreeHashHeaders patch above to see why.
                body: joinReadStreamWithPartDetails(fs_1.createReadStream(path, { start, end }), part),
                range: `bytes ${start}-${end}/*`,
                uploadId: uploadId,
                vaultName: s.vaultName,
            }).promise();
            return Buffer.from(assert_1.assertExists(res.checksum), "hex");
        });
    }
};
//# sourceMappingURL=AWSS3Glacier.js.map