"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const fs_1 = require("fs");
const http_1 = require("../util/http");
const sha1File_1 = require("../util/sha1File");
class AuthenticatedAPI {
    constructor(accountId, applicationKey) {
        // Promises returned are resolved/rejected at once.
        this.renewSync = [];
        // It's possible for these values to change due to authentication expiry and renewal.
        this._authToken = "";
        this._url = "";
        this._recommendedPartSize = -1;
        this.accountId = accountId;
        this.applicationKey = applicationKey;
    }
    // Ensure that multiple calls to renew() only causes one request and all
    get authToken() {
        return this._authToken;
    }
    get url() {
        return this._url;
    }
    get recommendedPartSize() {
        return this._recommendedPartSize;
    }
    renew() {
        return new Promise((resolve, reject) => {
            if (this.renewSync.push({ resolve, reject }) > 1) {
                return;
            }
            const basicAuth = Buffer.from(`${this.accountId}:${this.applicationKey}`).toString(`base64`);
            http_1.http({
                method: `GET`,
                url: `https://api.backblazeb2.com/b2api/v1/b2_authorize_account`,
                headers: {
                    Accept: `application/json`,
                    Authorization: `Basic ${basicAuth}`,
                },
            })
                .then(res => {
                this._authToken = res.authorizationToken;
                this._url = res.apiUrl;
                this._recommendedPartSize = res.recommendedPartSize;
                for (const { resolve } of this.renewSync) {
                    resolve();
                }
            }, err => {
                for (const { reject } of this.renewSync) {
                    reject(err);
                }
            })
                .then(() => {
                this.renewSync = [];
            });
        });
    }
}
exports.parseBackblazeB2Options = (args) => {
    const accountId = args.account.toString();
    const applicationKey = args.key.toString();
    const bucketId = args.bucket.toString();
    return {
        accountId, applicationKey, bucketId,
    };
};
const getUploadPartUrl = (apiUrl, authToken, uploadId) => __awaiter(this, void 0, void 0, function* () {
    const res = yield http_1.http({
        method: `POST`,
        url: `${apiUrl}/b2api/v1/b2_get_upload_part_url`,
        headers: {
            Accept: `application/json`,
            Authorization: authToken,
            "Content-Type": `application/json`,
        },
        body: {
            fileId: uploadId,
        },
    });
    return {
        url: res.uploadUrl,
        token: res.authorizationToken,
    };
});
exports.BackblazeB2 = {
    minParts: 2,
    maxParts: 10000,
    minPartSize: 1024 * 1024 * 5,
    maxPartSize: 1024 * 1024 * 1024 * 5,
    fromOptions(options) {
        return __awaiter(this, void 0, void 0, function* () {
            const api = new AuthenticatedAPI(options.accountId, options.applicationKey);
            yield api.renew();
            return {
                api,
                bucketId: options.bucketId,
            };
        });
    },
    completeUpload(s, uploadId, _, partHashes) {
        return __awaiter(this, void 0, void 0, function* () {
            return http_1.http({
                method: `POST`,
                url: `${s.api.url}/b2api/v1/b2_finish_large_file`,
                headers: {
                    Accept: `application/json`,
                    Authorization: s.api.authToken,
                    "Content-Type": `application/json`,
                },
                body: {
                    fileId: uploadId,
                    partSha1Array: partHashes.map(h => h.toString("hex")),
                },
            });
        });
    },
    idealPartSize(s, fileSize) {
        return __awaiter(this, void 0, void 0, function* () {
            // Use recommended part size provided by server if it is able to split the file into a valid amount of parts.
            if (Math.ceil(fileSize / s.api.recommendedPartSize) <= this.maxParts) {
                return s.api.recommendedPartSize;
            }
            return Math.ceil(fileSize / this.maxParts);
        });
    },
    initiateNewUpload(s, fileName, _) {
        return __awaiter(this, void 0, void 0, function* () {
            const res = yield http_1.http({
                method: `POST`,
                url: `${s.api.url}/b2api/v1/b2_start_large_file`,
                headers: {
                    Accept: `application/json`,
                    Authorization: s.api.authToken,
                    "Content-Type": `application/json`,
                },
                body: {
                    fileName,
                    bucketId: s.bucketId,
                    contentType: `application/octet-stream`,
                },
            });
            return res.fileId;
        });
    },
    uploadPart(s, uploadId, { path, number, start, end }) {
        return __awaiter(this, void 0, void 0, function* () {
            let uploadUrl;
            let authorizationToken;
            try {
                const res = yield getUploadPartUrl(s.api.url, s.api.authToken, uploadId);
                uploadUrl = res.url;
                authorizationToken = res.token;
            }
            catch (err) {
                if (err instanceof http_1.HTTPBadStatusError && err.statusCode === 401) {
                    // Failed to get upload URL due to authorisation, reauthorise.
                    yield s.api.renew();
                }
                // Still fail, even if had to renew and was successful.
                throw err;
            }
            const hash = yield sha1File_1.sha1File(fs_1.createReadStream(path, { start, end }));
            yield http_1.http({
                method: `POST`,
                url: uploadUrl,
                headers: {
                    Accept: `application/json`,
                    Authorization: authorizationToken,
                    "X-Bz-Part-Number": number + 1,
                    "Content-Length": end - start + 1,
                    "X-Bz-Content-Sha1": hash.toString("hex"),
                },
                body: fs_1.createReadStream(path, { start, end }),
            });
            return hash;
        });
    }
};
//# sourceMappingURL=BackblazeB2.js.map