"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const filesize_1 = __importDefault(require("filesize"));
const p_queue_1 = __importDefault(require("p-queue"));
const path_1 = __importDefault(require("path"));
const wait_1 = require("../util/wait");
const state_1 = require("./state");
// 5 minutes.
const MAX_RETRY_DELAY = 60 * 5;
exports.upload = (ctx, svc, s) => __awaiter(this, void 0, void 0, function* () {
    ctx.log(`Upload started`);
    ctx.updateProgress({
        description: `Checking for existing work...`,
        completeRatio: 0,
    });
    const { uploadId, filePath, fileLastChanged, partSize, partsNeeded } = (yield ctx.resumeSession()) ||
        (yield (() => __awaiter(this, void 0, void 0, function* () {
            ctx.updateProgress({
                description: `No existing upload found, initiating new upload...`,
                completeRatio: 0,
            });
            // Calculate appropriate part size.
            // Note that idealPartSize() can be non-deterministic.
            const partSize = Math.max(svc.minPartSize, Math.min(svc.maxPartSize, yield svc.idealPartSize(s, ctx.file.size)));
            // Validate how many parts are needed.
            const partsNeeded = Math.ceil(ctx.file.size / partSize);
            if (partsNeeded < 1) {
                throw new Error(`File is too small`);
            }
            if (partsNeeded > svc.maxParts) {
                throw new Error(`File is too big (${filesize_1.default(ctx.file.size)}) and cannot fit into ${svc.maxParts} ${filesize_1.default(svc.maxPartSize)} parts ` +
                    `(requires ${partsNeeded} parts)`);
            }
            // Initiate new upload with service.
            const uploadId = yield svc.initiateNewUpload(s, path_1.default.basename(ctx.file.path), partSize);
            // Build and save new session.
            const session = {
                uploadId,
                filePath: ctx.file.path,
                fileLastChanged: ctx.file.lastModified,
                partSize,
                partsNeeded,
            };
            yield ctx.writeSession(session);
            return session;
        }))());
    if (!ctx.force) {
        // Ensure that session uses same file with same contents and part size.
        if (filePath !== ctx.file.path || fileLastChanged !== ctx.file.lastModified) {
            throw new Error(`Cannot resume upload as file has changed since last session`);
        }
    }
    // TODO Add ability to verify integrity of hashes and invalidate specific parts.
    const partHashes = yield Promise.all(new Array(partsNeeded)
        .fill(0)
        .map((_, part) => state_1.loadPartHash(ctx, part)));
    let partsCompleted = partHashes.filter(h => h != null).length;
    // This value is used as the exponent for exponential backoff, and is
    // incremented for every part upload failure and reset to zero on every
    // successful upload.
    //
    // This means that intermittent failures will not encounter much delay (as
    // this value doesn't increase much and will quickly be reset), but service
    // issues will not cause constant rapid retries and wait longer until
    // the service has resumed normal performance.
    //
    // The maximum retry delay is defined in the constant MAX_RETRY_DELAY,
    // regardless of what 2 ^ $consecutiveFailures is.
    let consecutiveFailures = 0;
    const queue = new p_queue_1.default({ concurrency: ctx.concurrentUploads });
    const updateUploadProgress = () => {
        const partsRemaining = partsNeeded - partsCompleted;
        ctx.updateProgress({
            description: `Uploading ${partsRemaining} ${filesize_1.default(partSize)} parts (${filesize_1.default(partsRemaining * partSize)})...`,
            completeRatio: Math.min(0.99, partsCompleted / partsNeeded),
        });
    };
    updateUploadProgress();
    const queuePartUploadTask = (part) => {
        const start = part * partSize;
        const end = Math.min(ctx.file.size, (part + 1) * partSize) - 1;
        // Code awaits for queue to go idle later, so no need to handle individual queued upload Promises.
        queue.add(() => __awaiter(this, void 0, void 0, function* () {
            if (consecutiveFailures) {
                // Exponential backoff if currently in a series of failures.
                yield wait_1.wait(Math.min(MAX_RETRY_DELAY, Math.pow(2, consecutiveFailures)));
            }
            let hash;
            try {
                hash = yield svc.uploadPart(s, uploadId, { path: ctx.file.path, number: part, start, end });
            }
            catch (err) {
                // Part upload failed, increase delay, log, and requeue.
                consecutiveFailures++;
                ctx.log(`Failed to upload part ${part} with error "${err}", retrying...`);
                queuePartUploadTask(part);
                return;
            }
            // Part successfully uploaded.
            ctx.log(`Uploaded part ${part}`, true);
            partsCompleted++;
            // Reset failure streak.
            consecutiveFailures = 0;
            partHashes[part] = hash;
            yield state_1.savePartHash(ctx, part, hash);
            updateUploadProgress();
        }));
    };
    for (const [part, hash] of partHashes.entries()) {
        if (hash) {
            // Hash exists so already previously uploaded.
            continue;
        }
        queuePartUploadTask(part);
    }
    // NOTE: The queue could change during uploads due to retry,
    // so it's not as simple as awaiting all initial upload Promises.
    yield queue.onIdle();
    ctx.updateProgress({
        description: "All parts have been uploaded, finalising...",
        completeRatio: 0.99,
    });
    yield svc.completeUpload(s, uploadId, ctx.file.size, partHashes);
    ctx.log(`File successfully uploaded`);
});
//# sourceMappingURL=upload.js.map